package cmsc420_f20;

import java.util.ArrayList;
import java.util.List;

/**
 * An implementation of weight-Balanced Jackhammer (BJ) trees. The jackhammer
 * tree is based on rebalancing itself by destroying (jackhammering) and
 * rebuilding subtrees that are out of balance. Each key-value pair is
 * associated with a positive numeric weight. The weight of an internal node p,
 * denoted weight(p), is defined to be the total weight of the entries in its
 * subtree. The balance ratio of a node is defined to be:
 * 
 * balanceRatio(p) = max(weight(p.left), weight(p.right))/weight(p).
 * 
 * Given alpha, 1/2 < alpha < 1, we say that a tree is alpha-balanced if every
 * node has a balance ratio of at most alpha.
 * 
 * Unfortunately, the balance ratio alone is not a good criterion for when to
 * rebuild trees. This is because a subtree may contain a single entry of very
 * high weight, and rebuilding such a subtree may have no effect on its
 * structure. Given beta, where 0 < beta < 2 alpha - 1, we define the max ratio
 * of a node p to be
 * 
 * maxRatio(p) = maxWt(p) / weight(p)
 * 
 * where maxWt(p) is the maximum weight of any single entry in p's subtree. We
 * will not rebuild a subtree rooted at p if maxRatio(p) is larger than beta.
 * 
 * After each modification operation (insert or delete), we check the nodes on
 * the search path, starting from the root. If we encounter a node (closest to
 * the root) that is not (alpha,beta)-balanced, the associated subtree is
 * "jackhammered" by replacing p with a weight-balanced tree containing the same
 * entries. The entire subtree rooted at this node is rebuilt to an ideally
 * balanced form.
 * 
 * A subtree is rebuilt as follows. We first enumerate the external nodes of the
 * subtree in sorted order. Let list[0..k-1] denote the resulting list. The list
 * is split as evenly as possible by weight into left and right sublists, and
 * trees are generated recursively from them. (See buildSubtree for details.)
 *
 * Generic Elements: The tree is parameterized by two types, Key and Value. The
 * first is used for searching and the second is simply stored.
 * 
 * Parameters: The tree's balance is determined by the values ALPHA and BETA. To
 * assure reasonable behavior, we require that 1/2 < ALPHA < 1, and 0 < BETA <
 * 2*ALPHA - 1.
 */

public class BJTree<Key extends Comparable<Key>, Value> {

	private final boolean DEBUG = false; // produce extra debugging output
	private final float ALPHA = 0.66667f; // maximum allowed balance ratio
	private final float BETA = 0.25f; // max ratio filter

	// =================================================================
	// =================================================================
	// Node processing
	//
	// There are two types of nodes, external and internal. All nodes
	// key values. External nodes also store the associated value.
	// Internal nodes also store left and right child pointers.
	// =================================================================
	// =================================================================

	/**
	 * A node of the tree. This is an abstract object, which is extended for
	 * internal and external nodes.
	 */
	private abstract class Node { // generic node type
		float weight; // the weight associated with this node
		float maxWt; // the max weight in this subtree

		Node(float weight, float maxWt) { // basic constructor
			this.weight = weight;
			this.maxWt = maxWt;
		}

		// -----------------------------------------------------------------
		// Standard dictionary operations
		// -----------------------------------------------------------------

		abstract Value find(Key x); // find key in subtree

		abstract Node insert(Key x, Value v, float weight) throws Exception; // insert key-value into subtree

		abstract Node delete(Key x) throws Exception; // delete key from subtree

		abstract ArrayList<String> getPreorderList(); // list entries in preorder

		// -----------------------------------------------------------------
		// Additional dictionary operations
		// -----------------------------------------------------------------

		abstract Value getMin(); // get the minimum key in subtree

		abstract Value getMax(); // get the maximum key in subtree

		abstract Value findUp(Key x); // find next that is greater or equal

		abstract Value findDown(Key x); // find next that is lesser or equal

		// -----------------------------------------------------------------
		// Utility functions
		// -----------------------------------------------------------------

		abstract Node rebalance(Key x); // check and if needed, rebalance subtree containing x

		abstract List<ExternalNode> makeExternList(); // make list of external nodes

		abstract String debugPrint(String prefix); // print for debugging

		abstract public String toString(); // for debugging
	}

	// =================================================================
	// =================================================================
	// Internal node
	// =================================================================
	// =================================================================

	/**
	 * An internal node of the tree. The associated key is just used for splitting
	 * left and right subtrees.
	 */
	private class InternalNode extends Node {
		final Key key; // the key
		Node left; // children
		Node right;

		/**
		 * Test whether key x belongs in left subtree.
		 */
		boolean keyGoesOnLeft(Key x) {
			return x.compareTo(key) < 0;
		}

		/**
		 * Basic constructor. The weight of the node is the sum of the weights of its
		 * left and right children.
		 */
		InternalNode(Key key, Node left, Node right) {
			super(0, 0);
			this.key = key;
			this.left = left;
			this.right = right;
			updateWeight(); // set weight and maxWt
		}

		/**
		 * Update the weight and maxWt of this node, based on the weight of the two
		 * subtrees. (This should be called whenever the left or right subtree is
		 * changed, or their weights may have changed.)
		 */
		void updateWeight() {
			weight = left.weight + right.weight;
			maxWt = Math.max(left.maxWt, right.maxWt);
		}

		// -----------------------------------------------------------------
		// Standard dictionary operations
		// -----------------------------------------------------------------

		/**
		 * Find key in internal node.
		 */
		Value find(Key x) {
			if (keyGoesOnLeft(x)) { // x goes in left subtree
				return left.find(x);
			} else { // x goes in right subtree
				return right.find(x);
			}
		}

		/**
		 * Insert into this subtree.
		 *
		 * This just passes the request to the appropriate child.
		 */
		Node insert(Key x, Value v, float w) throws Exception {
			if (keyGoesOnLeft(x)) { // x goes in left subtree
				left = left.insert(x, v, w);
			} else { // x goes in right subtree
				right = right.insert(x, v, w);
			}
			updateWeight(); // update this node's weight
			return this;
		}

		/**
		 * Delete from this subtree.
		 * 
		 * Special processing is needed when the deleted node is an external node. This
		 * is signaled by the fact that the external delete function returns null. If
		 * so, we return a pointer to the "other" child.
		 */
		Node delete(Key x) throws Exception {
			if (keyGoesOnLeft(x)) { // delete from left
				left = left.delete(x);
				if (left == null)
					return right; // subtree is gone, return sibling
				else {
					updateWeight(); // else update our weight
					return this;
				}
			} else { // delete from right
				right = right.delete(x);
				if (right == null)
					return left; // subtree is gone, return sibling
				else {
					updateWeight(); // else update our weight
					return this;
				}
			}
		}

		/**
		 * Get a list of the nodes in preorder.
		 *
		 * Adds current key in parentheses, followed by left and right subtrees.
		 */
		ArrayList<String> getPreorderList() {
			ArrayList<String> list = new ArrayList<String>();
			list.add(toString() + " wt: " + weight); // add this node
			list.addAll(left.getPreorderList()); // add left
			list.addAll(right.getPreorderList()); // add right
			return list;
		}

		// -----------------------------------------------------------------
		// Other dictionary operations
		// -----------------------------------------------------------------

		/**
		 * Get the value associated with the minimum key.
		 *
		 */
		Value getMin() {
			return left.getMin();
		}

		/**
		 * Get the value associated with the maximum key.
		 *
		 */
		Value getMax() {
			return right.getMax();
		}

		/**
		 * Find next value with key greater or equal.
		 */
		Value findUp(Key x) {
			if (keyGoesOnLeft(x)) { // x goes in left subtree
				Value v = left.findUp(x);
				if (v == null)
					return right.getMin();
				else
					return v;
			} else { // x goes in right subtree
				return right.findUp(x);
			}
		}

		/**
		 * Find next value with key lesser or equal.
		 */
		Value findDown(Key x) {
			if (keyGoesOnLeft(x)) { // x goes in left subtree
				return left.findDown(x);
			} else { // x goes in right subtree
				Value v = right.findDown(x);
				if (v == null)
					return left.getMax();
				else
					return v;
			}
		}

		// -----------------------------------------------------------------
		// Utility functions
		// -----------------------------------------------------------------

		/**
		 * Determine whether a node needs to be rebuilt. Happens if:
		 * 
		 * max(left.weight,right.weight) / weight > alpha
		 *
		 * -and-
		 *
		 * maxWt / weight < beta
		 *
		 */
		boolean needsRebuilding() {
			float balanceRatio = Math.max(left.weight, right.weight) / weight;
			float maxRatio = maxWt / weight;
			if (DEBUG) {
				System.out.print(
						"...Checking node " + key + " balance-ratio = " + balanceRatio + " max-ratio = " + maxRatio);
				if ((balanceRatio > ALPHA) && (maxRatio <= BETA)) {
					System.out.println(" !!! Fails balance conditions !!!");
				} else {
					System.out.println(" is balanced");
				}
			}
			return (balanceRatio > ALPHA) && (maxRatio <= BETA);
		}

		/**
		 * Checks whether the subtree containing x needs to be rebuilt. If it does, we
		 * invoke rebuilt here. If it is balanced, we recurse on the child containing x.
		 * We return a pointer to the (possibly modified) subtree.
		 */
		Node rebalance(Key x) {
			if (needsRebuilding()) { // do we need to be rebuilt?
				if (DEBUG) {
					System.out.println("...Rebuilding subtree at node " + key);
				}
				List<ExternalNode> list = makeExternList(); // collect leaves
				Node rebuilt = buildSubtree(list); // rebuild this subtree
				if (DEBUG) {
					System.out.println("...Rebuilt subtree:" + System.lineSeparator() + rebuilt.debugPrint("..."));
				}
				return rebuilt;
			} else { // this node is balanced, check the child containing x
				if (keyGoesOnLeft(x)) { // x goes in left subtree
					left = left.rebalance(x);
				} else { // x goes in right subtree
					right = right.rebalance(x);
				}
				return this; // return pointer to this subtree
			}
		}

		/**
		 * Builds a list of external nodes in inorder.
		 */
		List<ExternalNode> makeExternList() {
			ArrayList<ExternalNode> list = new ArrayList<ExternalNode>();
			list.addAll(left.makeExternList()); // add left
			list.addAll(right.makeExternList()); // add right
			return list;
		}

		/**
		 * Debug print subtree.
		 *
		 * This prints the subtree for debugging. I prefer to print in reverse inorder,
		 * since the tree's structure is evident by turning the printed picture 90
		 * degrees clockwise.
		 */
		String debugPrint(String prefix) {
			return left.debugPrint(prefix + "| ") + System.lineSeparator() + prefix + toString() + " wt: " + weight
					+ " maxWt: " + maxWt + System.lineSeparator() + right.debugPrint(prefix + "| ");
		}

		/**
		 * Print node contents (for debugging).
		 */
		public String toString() {
			return "(" + key.toString() + ")";
		}
	}

	// =================================================================
	// =================================================================
	// External node
	// =================================================================
	// =================================================================

	/**
	 * An external node of the tree. This stores adds a value.
	 */
	private class ExternalNode extends Node {

		final Key key; // the key
		final Value value; // the associated value object

		/**
		 * Test whether key x belongs in left subtree.
		 */
		boolean keyGoesOnLeft(Key x) {
			return x.compareTo(key) < 0;
		}

		/**
		 * Basic constructor.
		 */
		ExternalNode(Key key, Value value, float weight) {
			super(weight, weight);
			this.key = key;
			this.value = value;
		}

		/**
		 * Copy constructor.
		 */
		ExternalNode(ExternalNode p) {
			super(p.weight, p.weight);
			this.key = p.key;
			this.value = p.value;
		}

		// -----------------------------------------------------------------
		// Standard dictionary operations
		// -----------------------------------------------------------------

		/**
		 * Find key in external node.
		 */
		Value find(Key x) {
			if (x.compareTo(key) == 0) // exists here now?
				return value;
			else
				return null;
		}

		/**
		 * Insertion at an external node.
		 *
		 * When x is smaller than the key, we create an internal node storing key, with
		 * x on the left and a copy of this node on the right. When x is greater or
		 * equal, we create an internal node storing x's key, with a copy of this node
		 * on the left and x on the right.
		 */
		Node insert(Key x, Value v, float w) throws Exception {
			if (x.compareTo(key) == 0) { // matches key?
				throw new Exception("Insertion of duplicate (or previously deleted) key");
			} else {
				ExternalNode q = new ExternalNode(x, v, w); // new external node
				InternalNode result = null;
				if (keyGoesOnLeft(x)) { // insert on left side (x must be smallest overall)
					result = new InternalNode(key, q, this);
				} else { // insert on right side (more typical)
					result = new InternalNode(x, this, q);
				}
				result.updateWeight(); // set the new node's weight values
				return result;
			}
		}

		/**
		 * Delete key from this node.
		 *
		 * We expect to find the deleted key here. If not, we throw an exception.
		 * Otherwise, we remove the node by returning null (see InternalNode deletion
		 * for more information).
		 */
		Node delete(Key x) throws Exception {
			if (x.compareTo(key) == 0) { // exists and matches?
				return null;
			} else {
				throw new Exception("Deletion of nonexistent (or previously deleted) key");
			}
		}

		/**
		 * Get preorder list. This just returns an encoding of the data in this node.
		 */
		ArrayList<String> getPreorderList() {
			ArrayList<String> list = new ArrayList<String>();
			list.add(toString()); // add this node
			return list;
		}

		// -----------------------------------------------------------------
		// Additional dictionary operations
		// -----------------------------------------------------------------

		/**
		 * Get the value associated with the minimum key.
		 */
		Value getMin() {
			return value;
		}

		/**
		 * Get the value associated with the maximum key.
		 */
		Value getMax() {
			return value;
		}

		/**
		 * Find next value with key greater or equal.
		 */
		Value findUp(Key x) {
			if (x.compareTo(key) <= 0)
				return value;
			else
				return null;
		}

		/**
		 * Find next value with key lesser or equal.
		 */
		Value findDown(Key x) {
			if (x.compareTo(key) >= 0)
				return value;
			else
				return null;
		}

		// -----------------------------------------------------------------
		// Utility functions
		// -----------------------------------------------------------------

		/**
		 * Rebalances node following an insertion or deletion. For external nodes,
		 * nothing needs to be done.
		 */
		Node rebalance(Key x) {
			return this;
		}

		/**
		 * Get list of external nodes.
		 */
		List<ExternalNode> makeExternList() {
			ArrayList<ExternalNode> list = new ArrayList<ExternalNode>();
			list.add(this); // add this node
			return list;
		}

		/**
		 * Debug print subtree.
		 */
		String debugPrint(String prefix) {
			return prefix + toString();
		}

		/**
		 * Convert to string (for debugging).
		 */
		public String toString() {
			return "[" + key.toString() + " " + value.toString() + "] wt: " + weight;
		}
	}

	// =================================================================
	// =================================================================
	// Private member data
	// =================================================================
	// =================================================================

	private Node root; // root of the tree

	// =================================================================
	// =================================================================
	// Public members
	// =================================================================
	// =================================================================

	/**
	 * Creates an empty tree.
	 */
	public BJTree() {
		root = null;
	}

	// -----------------------------------------------------------------
	// Standard dictionary operations
	// -----------------------------------------------------------------

	/**
	 * Find value associated with a given key.
	 *
	 * @param x The key being sought
	 * @return A reference to the associated value
	 */
	public Value find(Key x) {
		if (root == null)
			return null;
		else
			return root.find(x);
	}

	/**
	 * Insert a key-value pair.
	 *
	 * @param x The key
	 * @param v The value
	 * @param w The entry's weight
	 * @throws Exception If they key already appears in the dictionary
	 */
	public void insert(Key x, Value v, float w) throws Exception {
		if (DEBUG) {
			System.out.println("...Inserting " + x + " wt: " + w);
		}
		if (root == null) { // insert the key
			root = new ExternalNode(x, v, w);
		} else {
			root = root.insert(x, v, w);
		}
		if (DEBUG) {
			System.out.println(
					"...Tree after insertion and before rebalancing " + x + System.lineSeparator() + debugPrint("..."));
		}
		root = root.rebalance(x); // rebalance after insertion

		if (DEBUG) {
			System.out.println("...Tree after rebalancing " + x + System.lineSeparator() + debugPrint("..."));
		}
	}

	/**
	 * Delete a key.
	 *
	 * @param x The key to be deleted
	 * @throws Exception If they key already appears in the dictionary
	 */
	public void delete(Key x) throws Exception {
		if (DEBUG) {
			System.out.println("Tree: Deleting " + x);
		}
		if (root == null) { // empty tree?
			throw new Exception("Deletion of nonexistent key");
		} else {
			root = root.delete(x); // delete the key
		}
		if (DEBUG) {
			System.out.println(
					"...Tree after deletion and before rebalancing " + x + System.lineSeparator() + debugPrint("..."));
		}
		if (root != null) {
			root = root.rebalance(x); // rebalance after deletion
		}

		if (DEBUG) {
			System.out.println("...Tree after rebalancing " + x + System.lineSeparator() + debugPrint("..."));
		}
	}

	/**
	 * Clear the tree, removing all entries.
	 *
	 */
	public void clear() {
		root = null; // clear tree by setting root to null
	}

	/**
	 * Get a list of entries in preorder
	 *
	 * @return ArrayList of string encoded items in preorder
	 */
	public ArrayList<String> getPreorderList() {
		if (root == null) { // empty tree?
			return new ArrayList<String>(); // return an empty list
		} else {
			return root.getPreorderList();
		}
	}

	// -----------------------------------------------------------------
	// Additional operations
	// -----------------------------------------------------------------

	/**
	 * Get value associated with minimum key.
	 *
	 * @return A reference to the associated value, or null if tree empty
	 */
	public Value getMin() {
		if (root == null)
			return null;
		else
			return root.getMin();
	}

	/**
	 * Get value associated with maximum key.
	 *
	 * @return A reference to the associated value, or null if tree empty
	 */
	public Value getMax() {
		if (root == null)
			return null;
		else
			return root.getMax();
	}

	/**
	 * Find next value with key greater or equal.
	 *
	 * @param x The key being sought
	 * @return A reference to the associated value
	 */
	public Value findUp(Key x) {
		if (root == null)
			return null;
		else
			return root.findUp(x);
	}

	/**
	 * Find next value with key lesser or equal.
	 *
	 * @param x The key being sought
	 * @return A reference to the associated value
	 */
	public Value findDown(Key x) {
		if (root == null)
			return null;
		else
			return root.findDown(x);
	}

	// -----------------------------------------------------------------
	// Local utility functions
	// -----------------------------------------------------------------

	/**
	 * Recursively rebuilds the subtree rooted at this node given a sorted list of
	 * external nodes and their weights. If the list has a single element, we return
	 * a single external node. Otherwise, letting k denote the list's length k, it
	 * computes the splitting point b, that minimizes the absolute value of the
	 * difference in weights between list[0..b-1] and list[b..k-1], with ties broken
	 * in favor of the smaller b value. It recursively builds subtrees for these
	 * sublists and joins them under a common node containing list[b].
	 * 
	 * We assume that weights are strictly positive but less than Float.MAX_VALUE,
	 * otherwise bad things will happen.
	 */

	protected Node buildSubtree(List<ExternalNode> list) {
		if (list.size() == 1) {
			return new ExternalNode(list.get(0)); // copy this node
		} else {
			float totalWeight = 0;
			for (int i = 0; i < list.size(); i++) { // get total weight in the list
				totalWeight += list.get(i).weight;
			}
			float leftWeight = 0; // weight on the left side
			float rightWeight = totalWeight; // weight on the right side
			int b = 0; // best splitting index
			float minDiff = rightWeight - leftWeight; // minimum difference so far
			for (int i = 0; i < list.size(); i++) {
				leftWeight += list.get(i).weight;
				rightWeight -= list.get(i).weight;
				float curDiff = Math.abs(rightWeight - leftWeight);
				if (curDiff < minDiff) { // found better split point?
					minDiff = curDiff; // save it
					b = i + 1; // split just after this point
				}
			}
			// Invariant: 0 < bestSplit < list.size
			// split and recurse
			Node left = buildSubtree(list.subList(0, b));
			Node right = buildSubtree(list.subList(b, list.size()));
			return new InternalNode(list.get(b).key, left, right);
		}
	}

	/**
	 * Print the tree for debugging purposes. The prefix is printed at the head of
	 * each line (indentation).
	 */
	protected String debugPrint(String prefix) {
		if (root != null)
			return root.debugPrint(prefix);
		else
			return null;
	}

}
